"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Mocker = /** @class */ (function () {
    function Mocker() {
        if (Mocker.instance) {
            return Mocker.instance;
        }
        Mocker.instance = this;
        this.mockConfigData = {};
    }
    /**
     * Set global mock data configuration.
     * @param {object} mockConfigData
     */
    Mocker.prototype.setMockData = function (mockConfigData) {
        this.mockConfigData = mockConfigData;
        return this;
    };
    /**
     * Add an mock item to global mock data configuration.
     * @param {string} key
     * @param {any} val
     */
    Mocker.prototype.addMockItem = function (key, val) {
        this.mockConfigData[key] = val;
        return this;
    };
    /**
     * Reset global mock data configuration.
     * @param {string} key
     * @param {any} val
     */
    Mocker.prototype.reset = function () {
        this.setMockData({});
        return this;
    };
    /**
     * Check specified mock item & add it to global mock data configuration.
     * @param {MockItemInfo} mockItem
     */
    Mocker.prototype.mock = function (mockItem) {
        if (!mockItem.url || (typeof mockItem.url !== 'string' && !(mockItem.url instanceof RegExp))) {
            return this;
        }
        mockItem.method = /^(get|post|put|patch|delete|any)$/i.test(mockItem.method || '')
            ? mockItem.method
            : 'any';
        mockItem.delay = isNaN(mockItem.delay) ? 0 : Math.max(0, +mockItem.delay);
        mockItem.status = /^[1-5][0-9][0-9]$/.test(mockItem.status) ? +mockItem.status : 200;
        mockItem.header = typeof mockItem.header === 'object' ? mockItem.header : {};
        var key = mockItem.url + "-" + mockItem.method;
        this.addMockItem(key, mockItem);
        return this;
    };
    /**
     * Make a mock item that matches an HTTP GET request.
     * @param {RegExp | String} url
     * @param {any} response
     * @param {number} delay
     * @param {number} status
     * @param {object} header
     */
    Mocker.prototype.get = function (url, response, delay, status, header) {
        if (delay === void 0) { delay = 0; }
        if (status === void 0) { status = 200; }
        if (header === void 0) { header = {}; }
        this.mock({ url: url, method: 'get', response: response, delay: delay, status: status, header: header });
        return this;
    };
    /**
     * Make a mock item that matches an HTTP POST request.
     * @param {RegExp | String} url
     * @param {any} response
     * @param {number} delay
     * @param {number} status
     * @param {object} header
     */
    Mocker.prototype.post = function (url, response, delay, status, header) {
        if (delay === void 0) { delay = 0; }
        if (status === void 0) { status = 200; }
        if (header === void 0) { header = {}; }
        this.mock({ url: url, method: 'post', response: response, delay: delay, status: status, header: header });
        return this;
    };
    /**
     * Make a mock item that matches an HTTP PUT request.
     * @param {RegExp | String} url
     * @param {any} response
     * @param {number} delay
     * @param {number} status
     * @param {object} header
     */
    Mocker.prototype.put = function (url, response, delay, status, header) {
        if (delay === void 0) { delay = 0; }
        if (status === void 0) { status = 200; }
        if (header === void 0) { header = {}; }
        this.mock({ url: url, method: 'put', response: response, delay: delay, status: status, header: header });
        return this;
    };
    /**
     * Make a mock item that matches an HTTP PATCH request.
     * @param {RegExp | String} url
     * @param {any} response
     * @param {number} delay
     * @param {number} status
     * @param {object} header
     */
    Mocker.prototype.patch = function (url, response, delay, status, header) {
        if (delay === void 0) { delay = 0; }
        if (status === void 0) { status = 200; }
        if (header === void 0) { header = {}; }
        this.mock({ url: url, method: 'patch', response: response, delay: delay, status: status, header: header });
        return this;
    };
    /**
     * Make a mock item that matches an HTTP DELETE request.
     * @param {RegExp | String} url
     * @param {any} response
     * @param {number} delay
     * @param {number} status
     * @param {object} header
     */
    Mocker.prototype.delete = function (url, response, delay, status, header) {
        if (delay === void 0) { delay = 0; }
        if (status === void 0) { status = 200; }
        if (header === void 0) { header = {}; }
        this.mock({ url: url, method: 'delete', response: response, delay: delay, status: status, header: header });
        return this;
    };
    /**
     * Make a mock item that matches an HTTP GET, POST, PUT, PATCH or DELETE request.
     * @param {RegExp | String} url
     * @param {any} response
     * @param {number} delay
     * @param {number} status
     * @param {object} header
     */
    Mocker.prototype.any = function (url, response, delay, status, header) {
        if (delay === void 0) { delay = 0; }
        if (status === void 0) { status = 200; }
        if (header === void 0) { header = {}; }
        this.mock({ url: url, method: 'any', response: response, delay: delay, status: status, header: header });
        return this;
    };
    /**
     * Check whether the specified request url matchs a defined mock item.
     * If a match were found, return mock meta information, otherwise a null is returned.
     * @param {string} reqUrl
     * @param {string} reqMethod
     */
    Mocker.prototype.matchMockItem = function (reqUrl, reqMethod) {
        var requestMethod = reqMethod || 'get';
        for (var key in this.mockConfigData) {
            try {
                var info = this.mockConfigData[key];
                if (info.disable === 'yes') {
                    continue;
                }
                var method = ("" + info.method).toLowerCase();
                if (method !== 'any' && method !== ("" + requestMethod).toLowerCase()) {
                    continue;
                }
                if (Array.isArray(info.regexp) && info.regexp.length === 2
                    && new RegExp(info.regexp[0], info.regexp[1]).test(reqUrl)) {
                    return info;
                }
                if ((info.url instanceof RegExp) && info.url.test(reqUrl)) {
                    return info;
                }
                if (reqUrl.indexOf(info.url) !== -1) {
                    return info;
                }
            }
            catch (e) { }
        }
        return null;
    };
    return Mocker;
}());
exports.default = Mocker;
//# sourceMappingURL=mocker.js.map